package org.steveshipway.blueprints;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.directive.*;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.nbt.*;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;

// Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
//Minecraft 1.8
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.state.BlockState;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.util.BlockPos;
import net.minecraftforge.common.config.Configuration;
import net.minecraftforge.common.util.Constants;

import org.steveshipway.blueprints.util18.*;

public class Portfolio {
	// Configuration - static, so it applies to all portfolios
	static public Configuration config;
	public static boolean quiet = false;
	
	// array of all blueprints
	HashMap<String, Blueprint> allBlueprints;
	HashMap<String, Blueprint> autoBlueprints;
	Logger logger;
	String path;
	
	// For undo 
	// Should this be static?  If it is not, we have to pass portfolio to the build function...
	public static Undo undo = new Undo(); 
	
	public Portfolio(String bppath, Logger logger, Configuration config) {
		// initialise
		allBlueprints = new HashMap<String, Blueprint>();
		autoBlueprints = new HashMap<String, Blueprint>();
		this.logger = logger;
		this.path = bppath;
		this.config = config;
		this.quiet = config.get(Configuration.CATEGORY_GENERAL, "quiet", false).getBoolean(false);
	}
	private Directive parse(String blueprint, String s, BufferedReader br) {
		Directive d = null;
    	while( s.matches("\\s*(#(.*))?")) { 
    		try {s = br.readLine();	} 
    		catch (IOException e) {s="NOOP";} 
    	}
		if( s.matches("(?i)\\s*call\\s(.*)") ) {
			d = new CallDirective(s,blueprint,this,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*desc(.*)") ) {
			d = new DescDirective(s,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*opt.*") ) {
			d = new OptionDirective(s,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*block\\s.*") ) {
			d = new BlockDirective(s,this.path,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*chest\\s.*") ) {
			d = new ChestDirective(s,this.path,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*spawner\\s.*") ) {
			d = new SpawnerDirective(s,this.path,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*area\\s.*") ) {
			d = new AreaDirective(s,this.path,this.logger);
			return d;
		}
		// chance directive is odd as it requires the next item to be read
		if( s.matches("(?i)\\s*chance\\s+\\d+.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*chance\\s*(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse CHANCE directive: "+s);
	    		return null;
	    	}
	    	int c = Integer.parseInt(m.group(1));
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("Second directive NOT FOUND after CHANCE");
				return null;
			}
	    	Directive dd = parse(blueprint,ss,br);
	    	if( dd == null ) {
	    		logger.error("Bad directive after CHANCE");
	    		return null;
	    	}
			d = new ChanceDirective(c,dd,this.logger);
			return d;
		}
		// random directive requires next N directives to be read
		if( s.matches("(?i)\\s*random\\s+\\d+.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*random\\s*(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse RANDOM directive: "+s);
	    		return null;
	    	}
	    	int c = Integer.parseInt(m.group(1));
	    	String ss;
	    	List dl = new LinkedList();
	    	Directive dd = null;
	    	while( c > 0 ) {
	    		c -= 1;
	    		try {
	    			ss = br.readLine();
	    		} catch (IOException e) {
	    			logger.error("Sufficient directives NOT FOUND after RANDOM");
	    			return null;
	    		}
	    		dd = parse(blueprint,ss,br);
	    		if( dd == null ) {
	    			logger.error("Bad directive after RANDOM");
	    			return null;
	    		}
	    		dl.add(dd);
	    	}
			d = new RandomDirective(dl,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*auto.*") ) {
			d = new AutoDirective(s,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*(null|noop).*") ) {
			d = new NullDirective(s);
			return d;
		}
		if( s.matches("(?i)\\s*(mob|entity).*") ) {
			d = new MobDirective(s,this.path,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*ifbiome.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*ifbiome\\s+(\\S+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse IFBIOME directive: "+s);
	    		return null;
	    	}
	    	String match = m.group(1);
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("True outcome directive NOT FOUND after IFBIOME");
				return null;
			}
	    	Directive da = parse(blueprint,ss,br);
	    	if( da == null ) {
	    		logger.error("Bad 'true' directive after IFBIOME");
	    		return null;
	    	}
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("False outcome directive NOT FOUND after IFBIOME");
				return null;
			}
	    	Directive db = parse(blueprint,ss,br);
	    	if( db == null ) {
	    		logger.error("Bad 'false' directive after IFBIOME");
	    		return null;
	    	}
			d = new IfBiomeDirective(match,da,db,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*ifblock.*") ) {
			Set blks = new HashSet<Block>();
	    	Pattern p = Pattern.compile("(?i)^\\s*ifblock\\s+(-?\\d+)[,\\s]+(-?\\d+)[,\\s]+(-?\\d+)[,\\s]+(\\S+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse IFBLOCK directive: "+s);
	    		return null;
	    	}
	    	BlockPos offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
	    	String blocknames = m.group(4);
	    	String blox[] = blocknames.split("\\s*,+\\s*");
	    	for( String blockname: blox ) {
		    	if( !blockname.contains(":") ) { blockname = "minecraft:"+blockname; }
		    	IBlockState bs = Util.getState(Block.getBlockFromName(blockname),0);
		    	if( bs == null ) {
		    		logger.error("Invalid block name in IFBLOCK directive: "+blockname);
		    		return null;
		    	}
		    	blks.add(bs.getBlock());
	    	}
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("True outcome directive NOT FOUND after IFBLOCK");
				return null;
			}
	    	Directive da = parse(blueprint,ss,br);
	    	if( da == null ) {
	    		logger.error("Bad 'true' directive after IFBLOCK");
	    		return null;
	    	}
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("False outcome directive NOT FOUND after IFBLOCK");
				return null;
			}
	    	Directive db = parse(blueprint,ss,br);
	    	if( db == null ) {
	    		logger.error("Bad 'false' directive after IFBLOCK");
	    		return null;
	    	}
			d = new IfBlockDirective(offset,blks,da,db,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*ify.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*ify\\s+(=|<|>)\\s*(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse IFY directive: "+s);
	    		return null;
	    	}
	    	char cmp = m.group(1).charAt(0);
	    	int  value = Integer.parseInt(m.group(2));
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("True outcome directive NOT FOUND after IFY");
				return null;
			}
	    	Directive da = parse(blueprint,ss,br);
	    	if( da == null ) {
	    		logger.error("Bad 'true' directive after IFY");
	    		return null;
	    	}
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("False outcome directive NOT FOUND after IFY");
				return null;
			}
	    	Directive db = parse(blueprint,ss,br);
	    	if( db == null ) {
	    		logger.error("Bad 'false' directive after IFY");
	    		return null;
	    	}
			d = new IfYDirective(cmp,value,da,db,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*iffacing.*") ) {
	    	Pattern p = Pattern.compile("(?i)^\\s*iffacing\\s+(\\d+)");
	    	Matcher m = p.matcher(s);
	    	if( ! m.find() ) {
	    		logger.error("Unable to parse IFFACING directive: "+s);
	    		return null;
	    	}
	    	int  value = Integer.parseInt(m.group(1));
	    	String ss;
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("True outcome directive NOT FOUND after IFFACING");
				return null;
			}
	    	Directive da = parse(blueprint,ss,br);
	    	if( da == null ) {
	    		logger.error("Bad 'true' directive after IFFACING");
	    		return null;
	    	}
			try {
				ss = br.readLine();
			} catch (IOException e) {
	    		logger.error("False outcome directive NOT FOUND after IFFACING");
				return null;
			}
	    	Directive db = parse(blueprint,ss,br);
	    	if( db == null ) {
	    		logger.error("Bad 'false' directive after IFFACING");
	    		return null;
	    	}
			d = new IfFacingDirective(value,da,db,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*msg\\s(.*)") ) {
			d = new MsgDirective(s,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*(pic(ture)?|img|image)\\s(.*)") ) {
			d = new PictureDirective(s,this.path, this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*frame.*") ) {
			d = new FrameDirective(s,this.logger);
			return d;
		}
		if( s.matches("(?i)\\s*break.*") ) {
			d = new BreakDirective(s);
			return d;
		}
		return null;
	}
	public Logger getlogger() {
		return this.logger;
	}
	private boolean loadFile( File f )  {
		// read in this File object
	    BufferedReader br;
	    String blueprint = null;
	    LinkedList steps = null;
	    String description = null;
	    Boolean visible = true;
	    int lineNo = 0;
		if( ! this.quiet ) { logger.info("["+f.getName()+"]: Parsing blueprint file "); }
		try {
			br = new BufferedReader(new FileReader(f));
		} catch (FileNotFoundException e) {
			logger.error("["+f.getName()+"]: Unable to open file: "+e.getMessage());
			return false;
		}
	    try {
	        String line;
	        while ((line = br.readLine()) != null) {
	        	// work through lines in the file
	        	lineNo += 1;
	        	if( line.matches("\\s*#(.*)")) { continue; }
	        	if( line.matches("\\s*")) { continue; }
	        	if( line.matches("(?i)\\s*end(.*)")) {
	        		if( blueprint == null ) {
	        			logger.error("["+f.getName()+":"+lineNo+"]: Unbalanced END");
		        		visible = true;
		        		description = "No description";
		        		if( steps != null ) { steps.clear(); }
	        			continue;
	        		}
	        		// create new blueprint object
	        		Blueprint bp = new Blueprint(blueprint,description,visible,steps,logger);
	        		allBlueprints.put( blueprint,bp );
	        		if( ! this.quiet ) {logger.info("["+f.getName()+":"+lineNo+"]: Created new plan name ["+blueprint+"]: "
	        			+ description + " with "+steps.size()+" steps"
	        			+ ( (bp.auto_chance>0)?(", chance " + Math.floor(bp.auto_chance*org.steveshipway.blueprints.lib.Constants.CHANCE_BASE)):"")
	        		); }
	        		if( bp.auto_chance > 0 ) {
	        			autoBlueprints.put( blueprint, bp);
	        		}
	        		// reset everything
	        		blueprint = null;
	        		visible = true;
	        		description = "No description";
	        		continue;
	        	} 
	        	Pattern p = Pattern.compile("^\\s*(public |private |)\\s*plan\\s+(\\S+)\\s*(.*)",Pattern.CASE_INSENSITIVE);
	        	Matcher m = p.matcher(line);
	        	if( m.find() ) {
	        		// a new definition!
	        		blueprint = m.group(2);
	        		description = m.group(3);
	        		if( m.group(1).matches("(?i)private.*")) {
	        			visible = false;
	        		}
	        		steps = new LinkedList();
	        		if( this.defines(blueprint) ) {
	        			logger.error("["+f.getName()+":"+lineNo+"]: Redeclaration of plan ["+blueprint+"] ");
	        			return false;
	        		}
	        		if( ! this.quiet ) {logger.info("["+f.getName()+":"+lineNo+"]: Starting new plan name ["+blueprint+"]");}
	        		continue;
	        	}
	        	Directive d = parse(blueprint,line,br);
	        	if(d != null) {
	        		if(steps != null ) { steps.add(d); }
	        	} else {
		        	logger.error("["+f.getName()+":"+lineNo+"] Unknown or badly formatted directive in file: "+line);
		        	// return false;
	        	}
	        }
	    } catch (IOException e) {
			logger.error("["+f.getName()+"]: Unable to read file: "+e.getMessage());
			return false;
		} finally {
	        try {
				br.close();
			} catch (IOException e) {
				// ignore
			}
	    }
	    if( blueprint != null ) {
	    	logger.warn("["+f.getName()+":"+lineNo+"]: Incomplete blueprint declaration ["+blueprint+"] ");
	    	return false;
	    }
		return true;
	}
	private void loadDir(File dir) {
		for (final File fileEntry : dir.listFiles()) {
			if(fileEntry.isDirectory()) {
				if( fileEntry.getName().startsWith(".") ) {	continue; } // not current or parent
				if( fileEntry.getName().matches("img|nbt") ) { continue; } // ignore special subdirs
				loadDir(fileEntry);
				continue; 
			}
		    if ( fileEntry.getName().endsWith(".cfg") ) { // legacy naming suffix
		        if( ! loadFile( fileEntry ) ) {
		    	    logger.error("["+fileEntry.getName()+"]: Unable to read blueprint file " + fileEntry.getPath());
		        }
		        continue;
		    }
		    if ( fileEntry.getName().endsWith(".bp") ) {  // blueprints
		    	File tmlFile = new File(fileEntry.getPath().replace(".bp",".tml")); // do not load bp file if tml file exists
		    	File scFile = new File(fileEntry.getPath().replace(".bp",".schematic")); // or if the schematic exists 
		    	if( ! tmlFile.exists() && ! scFile.exists() ) { 
		    		if( ! loadFile( fileEntry ) ) {
		    	    	logger.error("["+fileEntry.getName()+"]: Unable to read blueprint file " + fileEntry.getPath());
		        	}
		    	}
		        continue;
		    }
		    if ( fileEntry.getName().endsWith(".tml")) { // ruins templates
		        if( ! Ruins.importTmlFile( this, fileEntry ) ) {
		    	    logger.error("["+fileEntry.getName()+"]: Unable to import Ruins template file " + fileEntry.getPath());
		        }
		        continue;
		    }
		    if ( fileEntry.getName().endsWith(".schematic")) {  // schematics
		        if( ! Schematic.importFile( this, fileEntry ) ) {
		    	    logger.error("["+fileEntry.getName()+"]: Unable to import Schematic file " + fileEntry.getPath());
		        }
		        continue;
		    }
		    logger.warn("Unrecognised file type in Blueprint directory: "+ fileEntry.getPath());
		}		
	}
	public void load() {
		// load all blueprints in from config directory 
		if( ! this.quiet ) {logger.info("Initialising portfolio from " + this.path);}
		allBlueprints.clear();
		autoBlueprints.clear();
		File folder = new File(this.path);
		loadDir(folder);
		if( ! this.quiet ) {
			logger.info("There are " + allBlueprints.size() + " blueprints defined");
			logger.info("There are " + autoBlueprints.size() + " autogenerating blueprints");
		}
	}
	// save the blueprint to the specified file
	public boolean save(String s, String f) {
		boolean rv = true;
		if( f == null ) { return false; }
		Blueprint bp = allBlueprints.get(s);
		if( bp == null ) { return false; }
		rv = bp.save(f);
		return rv;
	}
	// check to see if the blueprint exists
	public boolean defines(String s) {
		if( s == null ) { return false; }
		return allBlueprints.containsKey(s);	
	}
	// fetch the blueprint
	public Blueprint blueprint(String s) {
		Blueprint bp = allBlueprints.get(s);
		return bp;
	}
	// build the blueprint
	public boolean build(World world, String s,BlockPos playerLocation, int r) {
		if( s == null ) {
			logger.error("Somehow received null blueprint name.  How did you call this?");
			return false;
		}
		if( world.isRemote ) {
			logger.error("Portfolio build called when World is remote?  Should never happen.");
			return false;
		}
		if( playerLocation == null ) {
			logger.error("Somehow received null build location.  That shouldn't be possible.");
			return false;
		}
		if( ! this.defines(s) ) {
			logger.error( "Build called for invalid blueprint name " + s);
			return false;
		}
		if( ! this.quiet ) {
			logger.info("Calling build for blueprint " + s + " at " + playerLocation.getX() + "," + playerLocation.getY() + "," + playerLocation.getZ() + " rotation " + r);
		}
		if( config.get(Configuration.CATEGORY_GENERAL, "autoCommit", true).getBoolean(true) ) {
			commit();
		}
		boolean rv = allBlueprints.get(s).build(world,playerLocation,r&3);
		return rv;
	}
	public List list(String s) {
		List l = new ArrayList();
		Set set = allBlueprints.entrySet();
	    Iterator i = set.iterator();
	    l.clear();
	    while(i.hasNext()) {
	    	Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( bp.isVisible() && bp.name.startsWith(s)) {
	        	l.add(bp.name);	      
	    	}
	    }
		return l;
	}
	public List list() {
		return this.list("");
	}
	public Iterator iterator() {
		return this.allBlueprints.keySet().iterator();
	}
	public Iterator auto_iterator() {
		return this.autoBlueprints.keySet().iterator();
	}
	public List help(Boolean showAll) {
		List l = new ArrayList();
		Set set = allBlueprints.entrySet();
	    Iterator i = set.iterator();
	    int c = 0;
	    while( i.hasNext()) {
            Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( bp.isVisible() ) { c += 1; }
	    }
	    i = set.iterator();
	    l.add("There are " + allBlueprints.size() + " blueprints defined ("+c+" public)");
	    while(i.hasNext()) {
	    	String s = new String();
            Map.Entry me = (Map.Entry)i.next();
            Blueprint bp = (Blueprint)me.getValue();
            if( showAll || bp.isVisible() ) {
              s = (bp.isVisible()?"":"[H]")+((bp.auto_chance>0)?"[A]":"") + bp.name + " : " + bp.helpText();
	          l.add(s);	         
            }
	    }
		return l;
	}
	public boolean delete(String s) {
		if( allBlueprints.containsKey(s) ) {
			allBlueprints.remove(s);
		}
		if( autoBlueprints.containsKey(s) ) { 
			autoBlueprints.remove(s);
		}
		return true;
	}
	// copy from the world
	public Blueprint copy(World world, String name, BlockPos from, BlockPos to, int opts ) {
		Blueprint bp = null;
		LinkedList steps = new LinkedList();
		int stepx, stepy, stepz;
		Set<Block> decorations = new HashSet<Block>();
		
		if( defines(name) ) {
			logger.error("Blueprint "+name+" already exists");
			return null;
		}
		
		// set up the list of decoration blocks for the second pass
		// these are blocks that fall off if not attached to something
		// so we do them in a 2nd pass -- mobs are done in the 3rd pass
		decorations.add(Blocks.rail); 
		decorations.add(Blocks.torch);
		decorations.add(Blocks.redstone_torch);	    		
		decorations.add(Blocks.redstone_wire);
		decorations.add(Blocks.wall_banner);
		decorations.add(Blocks.standing_banner);
		decorations.add(Blocks.carpet);
		decorations.add(Blocks.trapdoor);
		decorations.add(Blocks.skull);
		decorations.add(Blocks.tripwire);
		decorations.add(Blocks.lever);
		decorations.add(Blocks.stone_button);
		decorations.add(Blocks.wooden_button);
		decorations.add(Blocks.yellow_flower); 
		decorations.add(Blocks.red_flower);
		decorations.add(Blocks.double_plant);
		decorations.add(Blocks.wheat);
		decorations.add(Blocks.pumpkin);
		decorations.add(Blocks.melon_block);
		decorations.add(Blocks.pumpkin_stem);
		decorations.add(Blocks.melon_stem);
		decorations.add(Blocks.waterlily);
		decorations.add(Blocks.reeds);
		decorations.add(Blocks.cactus);
		decorations.add(Blocks.potatoes);
		decorations.add(Blocks.tallgrass);

		// iterate over the entire area.  Make sure coordinates are in correct order.
		if( from.getX() <= to.getX() ) { stepx = 1; } else { stepx = -1; }
		if( from.getY() <= to.getY() ) { stepy = 1; } else { stepy = -1; }
		if( from.getZ() <= to.getZ() ) { stepz = 1; } else { stepz = -1; }
		int curDv;
		Block curBlock;
		BlockPos offset = null;
		NBTTagCompound nbt = null;
		TileEntity te = null;
		String nbtStr;
		BlockPos blk;
		IBlockState iblockstate;
		// identify configuration defaults
		boolean copyAir = this.config.get("blueprint", "copyAir", "false").getBoolean(false);
		boolean copyLiquids = this.config.get("blueprint", "copyLiquids", "false").getBoolean(false);
			
		for( int pass = 0; pass < 2; pass += 1 ) {
			for( int x = from.getX(); x != (to.getX()+stepx); x += stepx ) {
				for( int y = from.getY(); y != (to.getY()+stepy); y += stepy ) {
					for( int z = from.getZ(); z != (to.getZ()+stepz); z += stepz ) {
						blk = new BlockPos(x,y,z);
						iblockstate = Util.getBlockState(world,blk);
						curBlock = iblockstate.getBlock();
						if( decorations.contains(curBlock) ^ (pass==1) )  {
							// only do decorations in step1, normal blocks in step 0
							continue;
						}
						if( (copyAir || (curBlock != Blocks.air)) 
							&& (copyLiquids || (curBlock != Blocks.water && curBlock != Blocks.lava && curBlock != Blocks.flowing_water && curBlock != Blocks.flowing_lava)) ) {
							//curDv = blockstate.getBlockMetadata(blk);
							curDv = Util.getDvFromState(iblockstate);
							offset = new BlockPos(x - from.getX(),y - from.getY(),z - from.getZ());
							te = Util.getTileEntity(world,blk);
							if(te != null) {
							//Block b = te.getBlockType();
								nbt = new NBTTagCompound();
								te.writeToNBT(nbt); // note this does not remove existing tags
								nbt.removeTag("x"); // remove the tags we don't need
								nbt.removeTag("y");
								nbt.removeTag("z");
								nbtStr = nbt.toString();							
							} else {
								nbtStr = null;
							}
							steps.add(new BlockDirective(curBlock,curDv,false,offset,nbtStr));
							//logger.info("Added block of type "+Block.blockRegistry.getNameForObject(curBlock)+" at position "+x+","+y+","+z+" ("+offset.posX+","+offset.posY+","+offset.posZ+")");
						} else {
							//logger.info("Omitting block of type "+Block.blockRegistry.getNameForObject(curBlock)+" at position "+x+","+y+","+z);
						} // if copyable
					} // z
				} // y
			} // x
		} // 2 pass
		// now scan for mobile entities in the area.  Expand slightly to get the ones hanging off the edge.
		// remember that this is a bound as doubles, but from/to are int, so need to raise upper bound by 1.0
		AxisAlignedBB aabb = Util.getBB(from,to);
		//logger.info("Check box for entities: "+aabb.toString());
		List<Entity> entityList = world.getEntitiesWithinAABBExcludingEntity((Entity)null,aabb);
		
		// These are mobs: Minecart, Painting, ArmourStand
		
		if ( this.config.get("blueprint", "copyMobs", "true").getBoolean(true) ) {
			
			for(Entity entity : entityList) {
				//logger.info("Found an entity: "+entity.getName()+" of class "+entity.getClass().getSimpleName()+" at "+entity.posX+","+entity.posZ);
				if( entity instanceof EntityPlayer ) { continue; }  // don't copy the player themselves
				nbt = new NBTTagCompound();
				entity.writeToNBT(nbt);
				nbt.setString("id",entity.getClass().getSimpleName().substring(6)); // class name minus the initial 'Entity'
				blk = new BlockPos( (int)Math.floor(entity.posX)-from.getX(),(int)Math.floor(entity.posY)-from.getY(),(int)Math.floor(entity.posZ)-from.getZ());
	
				// remove unnecessary tags
				nbt.removeTag("Pos");  // mobs
				nbt.removeTag("HurtByTimestamp");
				nbt.removeTag("PortalCooldown");
				nbt.removeTag("AbsorbtionAmount");
				nbt.removeTag("FallDistance");
				nbt.removeTag("DeathTime");
				nbt.removeTag("Motion");
				nbt.removeTag("UUIDLeast");
				nbt.removeTag("UUIDMost");
				nbt.removeTag("Air");
				nbt.removeTag("Dimension");
				//nbt.removeTag("Rotation");
				nbt.removeTag("File");
				nbt.removeTag("HurtTime");
				nbt.removeTag("TileX"); // paintings
				nbt.removeTag("TileY");
				nbt.removeTag("TileZ");
				steps.add(new MobDirective(blk,nbt));
			}
		}
		
		// create blueprint object ( no description, visible )
		bp = new Blueprint(name,"Copied from world",true,steps,logger);
		allBlueprints.put(name, bp);
		return bp;
	}
	// undo functions
	public void commit() {
		logger.info("Clearing out undo buffer.");
		this.undo.clear();
	}
	public void rollback(World world) {
		logger.warn("Rolling back all changes in the Undo buffer...");
		this.undo.rollback(world);
	}
	public int buffer_size() {
		return this.undo.size();
	}
}
