package org.steveshipway.blueprints;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.nbt.CompressedStreamTools;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagByteArray;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagDouble;
import net.minecraft.nbt.NBTTagList;
import net.minecraftforge.common.util.Constants;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.directive.*;

// Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
// Minecraft 1.8
import net.minecraft.util.BlockPos;
import org.steveshipway.blueprints.util18.*;

// This should read in Schematic format http://minecraft.gamepedia.com/Schematic

public class Schematic {

	public static boolean importFile(Portfolio portfolio,File fileEntry) {
		Blueprint bp;
		String name = "schm";
		String description = "Imported from Schematic "+fileEntry.getName();
		LinkedList steps = new LinkedList();
		LinkedList decorations = new LinkedList();
		int width, length, height;
		String cvtFileName = fileEntry.getName()+".bp";
		NBTTagCompound nbtTag;
		Writer writer = null;
    	Directive d;
    	Hashtable tileEntities = new Hashtable();
		Block block;
		int id, dv;
		boolean softAir = true;
		boolean centrePlan = true;
		
		if( ! portfolio.quiet ) { portfolio.logger.info("["+fileEntry.getName()+"]: Parsing Schematic file "); }

		softAir = portfolio.config.get("schematic", "softAir", true).getBoolean(true);
		centrePlan = portfolio.config.get("schematic", "centre", true).getBoolean(true);
		
		// Create conversion file
		name = fileEntry.getName().replace(".schematic","");
		cvtFileName = fileEntry.getPath().replace(".schematic",".bp");

		try {
			nbtTag = CompressedStreamTools.readCompressed(new FileInputStream(fileEntry));
		} catch (IOException e) {
			portfolio.logger.error("Unable to read NBT from Schematic file "+fileEntry.getName() +" : "+e.toString());
			return false;
		}
	
		// open conversion file
		try {
		    writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream(cvtFileName), "utf-8")
		    );
		    writer.write("# Blueprint for: "+ name +"\n");
		    writer.write("# Converted from schematic file "+fileEntry.getName()+"\n");
		    writer.write("\n");
		} catch (IOException ex) {
			portfolio.logger.error("Unable to write file "+cvtFileName+": "+ex.getMessage());
			try {writer.close();} catch(Exception ex2) {}
			writer = null;
		}
		
		// Convert to Blueprint
		if(!nbtTag.hasKey("Width",Constants.NBT.TAG_SHORT) || !nbtTag.hasKey("Blocks",Constants.NBT.TAG_BYTE_ARRAY)) {
			portfolio.logger.error("["+fileEntry.getName()+"]: File NBT does not appear to be in correct format");
			return false;
		}
		width  = nbtTag.getInteger("Width");
		height = nbtTag.getInteger("Height");
		length = nbtTag.getInteger("Length");
		
		// First, index the tile entities, so they can be referenced by the blocks
    	NBTTagList teList = nbtTag.getTagList("TileEntities",Constants.NBT.TAG_COMPOUND);
    	if( teList != null ) {
	    	for (int i = 0; i < teList.tagCount(); ++i) {
	    		NBTTagCompound te = (NBTTagCompound) teList.getCompoundTagAt(i);
	    		int x = te.getInteger("x");
	    		int y = te.getInteger("y");
	    		int z = te.getInteger("z");
	    		tileEntities.put(""+x+","+y+","+z,te.copy());
	    	}
		}
		
		// Process all blocks in the array
    	byte[] blocks = nbtTag.getByteArray("Blocks");
    	byte[] data   = nbtTag.getByteArray("Data");
       	for (int j = 0; j < height; ++j) {  // do y outside, so we build from bottom up
           	for (int i = 0; i < width; ++i) {
            	for (int k = 0; k < length; ++k) {
			    	int offset = (j*length + k)*width + i;
			    	String nbtstr = null;
			    	id = blocks[offset] & 255;
					dv = data[offset] & 15;
					if( tileEntities.containsKey(""+i+","+j+","+k)) {
						nbtstr = tileEntities.get(""+i+","+j+","+k).toString();
					}
					if( (id == 0) && softAir ) { continue; }
					block = Block.getBlockById(id);
					if( block != null ) {
						d = new BlockDirective(block,dv,false,new BlockPos(i,j,k),nbtstr); 
						if(id==50||id==75||id==76||id==96||id==177||id==68||id==69||id==77||id==143) { // torch, redstone torch, trapdoor, banner, wallsign, lever, button.  Item frames are entities.
							decorations.add(d);
						} else {
							steps.add(d);
						}	
					} else {
						portfolio.logger.warn("Unrecognised block ID: "+id);
					}
            	}
        	}
    	}
    	// Append the decorations steps to the others (this helps them attach properly)
		Iterator di = decorations.iterator();
	    while(di.hasNext()) { steps.add(di.next());  }

		
		// Now add any mobile entities
    	NBTTagList meList = nbtTag.getTagList("Entities",Constants.NBT.TAG_COMPOUND);
    	if( meList != null ) {
	    	for (int i = 0; i < meList.tagCount(); ++i) {
	    		NBTTagCompound me = (NBTTagCompound) meList.getCompoundTagAt(i);
	    		NBTTagList pos = me.getTagList("Pos",Constants.NBT.TAG_DOUBLE);
	    		int x = (int) Math.floor(Util.getDoubleAt(pos, 0));
	    		int y = (int) Math.floor(Util.getDoubleAt(pos, 1)+0.5);
	    		int z = (int) Math.floor(Util.getDoubleAt(pos, 2));
	    		steps.add( new MobDirective(new BlockPos(x,y,z),me));
	    	}
		}
		
		// Create the new blueprint
		bp = new Blueprint("_"+name,description,false,steps,portfolio.logger);
		
		portfolio.allBlueprints.put( "_"+name,bp );
		if( ! portfolio.quiet ) { portfolio.logger.info("["+fileEntry.getName()+"]: Created new blueprint name ["+name+"] from Schematic size (x,y,z)=("+width+","+length+","+height+")"); }
				
		// write new blueprint, and close conversion file
		if( writer != null ) {
		    try { 
		    	writer.write("\n# Main plan\n");
		    	writer.write(bp.asText()+"\n");
		    	writer.write("# End of blueprint\n");
			} catch (IOException ex) {
				portfolio.logger.error("Unable to write to file "+cvtFileName+": "+ex.getMessage());
			}
		}
		
		// Now create the main offset call
		steps = new LinkedList();
		
		if( nbtTag.hasKey("WEOffsetX") ) { 
			steps.add( new CallDirective(bp,new BlockPos(nbtTag.getInteger("WEOffsetX"),nbtTag.getInteger("WEOffsetY"),nbtTag.getInteger("WEOffsetZ")),0));
		} else if ( centrePlan ) {
			steps.add( new CallDirective(bp,new BlockPos(-(int)Math.floor(width/2),0,0),0));			
		} else {
			steps.add( new CallDirective(bp,new BlockPos(0,0,0),0));			
		}
		bp = new Blueprint(name,description,true,steps,portfolio.logger);
		portfolio.allBlueprints.put( name,bp );
		if( writer != null ) {
		    try { 
		    	writer.write("\n# Offset plan\n");
		    	writer.write(bp.asText()+"\n");
		    	writer.write("# End of blueprint\n");
			} catch (IOException ex) {
				portfolio.logger.error("Unable to write to file "+cvtFileName+": "+ex.getMessage());
			}
		}
		
		if( writer != null ) {
			try {writer.close();} catch (Exception ex) {}
		}		
		return true; 
	}
	
} 

